import os

from renard.cli import main


def test_nearest(capfd):
    code = main("nearest R10 21".split())
    out, err = capfd.readouterr()
    assert code == os.EX_OK
    assert out == "20\n"


def test_nearest_with_symbol(capfd):
    code = main("nearest R10 21000 -s".split())
    out, err = capfd.readouterr()
    assert code == os.EX_OK
    assert out == "20 k\n"


def test_nearby(capfd):
    code = main("nearby R20 31".split())
    out, err = capfd.readouterr()
    assert code == os.EX_OK
    assert out == "28\n31.5\n35.5\n"


def test_gt(capfd):
    code = main("gt R20 31".split())
    out, err = capfd.readouterr()
    assert code == os.EX_OK
    assert out == "31.5\n"


def test_lt(capfd):
    code = main("lt R20 31".split())
    out, err = capfd.readouterr()
    assert code == os.EX_OK
    assert out == "28\n"


def test_ge(capfd):
    code = main("ge R20 40".split())
    out, err = capfd.readouterr()
    assert code == os.EX_OK
    assert out == "40\n"


def test_le(capfd):
    code = main("le R20 40".split())
    out, err = capfd.readouterr()
    assert code == os.EX_OK
    assert out == "40\n"


def test_series_r3(capfd):
    code = main("series R5".split())
    out, err = capfd.readouterr()
    assert code == os.EX_OK
    assert out == "1.0\n1.6\n2.5\n4.0\n6.3\n"


def test_range_R10(capfd):
    code = main("range R10 1700 3400".split())
    out, err = capfd.readouterr()
    assert code == os.EX_OK
    assert out == "2e3\n2.5e3\n3.15e3\n"


def test_range_R10_symbol(capfd):
    code = main("range R10 1700 3400 -s".split())
    out, err = capfd.readouterr()
    assert code == os.EX_OK
    assert out == "2 k\n2.5 k\n3.15 k\n"


def test_bogus_r_series_gives_exit_code_ex_dataerr():
    code = main("series R13".split())
    assert code == os.EX_DATAERR


def test_bogus_value_gives_exit_code_ex_dataerr():
    code = main("nearest R10 FOO".split())
    assert code == os.EX_DATAERR


def test_malformed_command_gives_code_ex_usage():
    code = main("foo R13 316".split())
    assert code == os.EX_USAGE


def test_precision(capfd):
    code = main("precision R5".split())
    out, err = capfd.readouterr()
    assert code == os.EX_OK
    assert out == '0.01\n'

def test_bogus_r_series_precision_gives_exit_code_ex_dataerr():
    code = main("series R13".split())
    assert code == os.EX_DATAERR



from math import floor, log10

from renard.renard import _round_sig

PREFIXES = 'yzafpnµm kMGTPEZY'


def eng_string(x, sig_figs=3, prefix=True):
    """
    Returns float/int value <x> formatted in a simplified engineering format -
    using an exponent that is a multiple of 3.

    Args:
        sig_figs: number of significant figures

    prefix: Use SI suffix for exponent, e.g. k instead of e3, n instead of
    e-9 etc.
    """
    x = float(x)
    sign = ''
    if x < 0:
        x = -x
        sign = '-'

    if x == 0:
        exp3 = 0
        x3 = 0
    else:
        exp = int(floor(log10(x)))
        exp3 = exp - (exp % 3)
        x3 = x / (10 ** exp3)
        x3 = _round_sig(x3, 3)  # None of the Renard series values have more than 3 s.f.
        if x3 == int(x3):  # prevent from displaying .0
            x3 = int(x3)

    if prefix and (-24 <= exp3 <= 24) and (exp3 != 0):
        exp3_text = ' ' + PREFIXES[exp3 // 3 + 8]
    elif exp3 == 0:
        exp3_text = ''
    else:
        exp3_text = 'e' + str(exp3)

    t3 = str(x3)

    return ''.join((sign, t3, exp3_text))


"""The command-line for renard"""

import os
import sys

import docopt
import docopt_subcommands as dsc

from renard.eng import eng_string
from renard.version import __version__
from renard.renard import (series_key_from_name, find_nearest, find_nearest_few, find_greater_than_or_equal,
                           find_greater_than, find_less_than, find_less_than_or_equal, series, rrange, precision)

DOC_TEMPLATE = """{program}

Usage: {program} [options] <command> [<args> ...]

Options:
  -h --help     Show this screen.
  -v --verbose  Use verbose logging

Available commands:
  {available_commands}

See '{program} help <command>' for help on specific commands.
"""


@dsc.command()
def handle_nearest(precommand, args):
    """usage: {program} nearest <Renard-series> <value> [--symbol]

    The nearest value in an Renard series.

    Options:
      -s --symbol  Use the SI magnitude prefix symbol.
    """
    series_key = extract_series_key(args)
    value = extract_value(args)
    nearest = find_nearest(series_key, value)
    nearest_text = present_value(args, nearest)
    print(nearest_text)
    return os.EX_OK


@dsc.command()
def handle_nearby(precommand, args):
    """usage: {program} nearby <Renard-series> <value> [--symbol]

    At least three nearby values in an Renard series, and least one of
    which will be less-than the given value, and at least one
    greater-than the given value.

    Options:
      -s --symbol  Use the SI magnitude prefix symbol.
    """
    series_key = extract_series_key(args)
    value = extract_value(args)
    nearby_few = find_nearest_few(series_key, value)
    for item in nearby_few:
        item_text = present_value(args, item)
        print(item_text)
    return os.EX_OK


@dsc.command()
def handle_gt(precommand, args):
    """usage: {program} gt <Renard-series> <value> [--symbol]

    The largest value greater-than the given value.

    Options:
      -s --symbol  Use the SI magnitude prefix symbol.
    """
    series_key = extract_series_key(args)
    value = extract_value(args)
    nearest = find_greater_than(series_key, value)
    nearest_text = present_value(args, nearest)
    print(nearest_text)
    return os.EX_OK


@dsc.command()
def handle_ge(precommand, args):
    """usage: {program} ge <Renard-series> <value> [--symbol]

    The largest value greater-than or equal-to the given value.

    Options:
      -s --symbol  Use the SI magnitude prefix symbol.
    """
    series_key = extract_series_key(args)
    value = extract_value(args)
    nearest = find_greater_than_or_equal(series_key, value)
    nearest_text = present_value(args, nearest)
    print(nearest_text)
    return os.EX_OK


@dsc.command()
def handle_lt(precommand, args):
    """usage: {program} lt <Renard-series> <value> [--symbol]

    The largest value less-than the given value.

    Options:
      -s --symbol  Use the SI magnitude prefix symbol.
    """
    series_key = extract_series_key(args)
    value = extract_value(args)
    nearest = find_less_than(series_key, value)
    nearest_text = present_value(args, nearest)
    print(nearest_text)
    return os.EX_OK


@dsc.command()
def handle_le(precommand, args):
    """usage: {program} le <Renard-series> <value> [--symbol]

    The largest value less-than or equal-to the given value.

    Options:
      -s --symbol  Use the SI magnitude prefix symbol.
    """
    series_key = extract_series_key(args)
    value = extract_value(args)
    nearest = find_less_than_or_equal(series_key, value)
    nearest_text = present_value(args, nearest)
    print(nearest_text)
    return os.EX_OK


@dsc.command()
def handle_series(precommand, args):
    """usage: {program} series <Renard-series>

    The base values for the given Renard series.
    """
    series_key = extract_series_key(args)
    for item in series(series_key):
        print(item)
    return os.EX_OK


@dsc.command()
def handle_range(precommand, args):
    """usage: {program} range <Renard-series> <start-value> <stop-value> [--symbol]

    All values in the given Renard series from start-value to stop-value inclusive.

    Options:
      -s --symbol  Use the SI magnitude prefix symbol.
    """
    series_key = extract_series_key(args)
    start_value = extract_value(args, '<start-value>')
    stop_value = extract_value(args, '<stop-value>')
    items = rrange(series_key, start_value, stop_value)
    for item in items:
        item_text = present_value(args, item)
        print(item_text)
    return os.EX_OK


@dsc.command()
def handle_precision(precommand, args):
    """usage: {program} precision <Renard-series>

    The multiple to which the base values in the series are rounded.
    """
    series_key = extract_series_key(args)
    p = precision(series_key)
    print(p)
    return os.EX_OK


def present_value(args, nearest):
    return eng_string(nearest, prefix=args['--symbol'])


def extract_series_key(args):
    e_series_name = args['<Renard-series>']
    series_key = series_key_from_name(e_series_name)
    return series_key


def extract_value(args, name='<value>'):
    text_value = args[name]
    try:
        value = float(text_value)
    except ValueError:
        raise ValueError("{!r} could not be interpreted as an Renard series {}".format(
            text_value, name[1:-1]))
    return value


def main(argv=None):
    try:
        return dsc.main(
            program='renard',
            argv=argv,
            doc_template=DOC_TEMPLATE,
            exit_at_end=False)
    except (docopt.DocoptExit, SystemExit) as exc:
        print(exc, file=sys.stderr)
        return os.EX_USAGE
    except ValueError as exc:
        print(exc, file=sys.stderr)
        return os.EX_DATAERR


if __name__ == '__main__':
    sys.exit(main())
